(*----------------------------------------------------------------------------*
 *  DirectShow sample from DirectX 8.0 SDK                                    *
 *  Delphi adaptation by Alexey Barkovoy                                      *
 *  E-Mail: clootie@reactor.ru                                                *
 *                                                                            *
 *  This sample is modified Texture3D sample from SDK, what uses DMO model    *
 *  as filter implementation. So it's also shows how to use helper class      *
 *  from DMOimpl.h                                                            *
 *  Note: DMO - DirectX Media Objects                                         *
 *                                                                            *
 *  Modified: 14-May-2002                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru                                                *
 *----------------------------------------------------------------------------*)

unit TextureRenderer;

interface                                 

uses
  Windows, ActiveX, ComObj, DirectShow,
  {$IFDEF DXG_COMPAT} DirectXGraphics {$ELSE} Direct3D8 {$ENDIF},
  D3DX8, DShow3DBridge;

type
  TTextureRenderer = class(TMediaObjectImpl, IMarshal,
    IMediaObject, IDirect3DDeviceExch)
  private
    fD3DDevice: IDirect3DDevice8;
    fTexture: IDirect3DTexture8;
    ///////////
    fTextureFormat: TD3DFormat;
    fIMarshal: IMarshal;
    fMediaBuffer: IMediaBuffer;
    fTimestamp,
    fTimelength: TREFERENCE_TIME;
    procedure InitD3DTexture;
  protected
    function InternalGetInputStreamInfo(dwInputStreamIndex: DWORD; out pdwFlags: DWORD): HRESULT; override;
    function InternalGetOutputStreamInfo(dwOutputStreamIndex: DWORD; out pdwFlags: DWORD): HRESULT; override;
    function InternalCheckInputType(dwInputStreamIndex: DWORD; const pmt: PDMO_MEDIA_TYPE): HRESULT; override;
    function InternalCheckOutputType(dwOutputStreamIndex: DWORD; const pmt: PDMO_MEDIA_TYPE): HRESULT; override;
    function InternalGetInputType(dwInputStreamIndex, dwTypeIndex: DWORD; out pmt: TDMO_MEDIA_TYPE): HRESULT; override;
    function InternalGetOutputType(dwOutputStreamIndex, dwTypeIndex:DWORD; out pmt: TDMO_MEDIA_TYPE): HRESULT; override;
    function InternalGetInputSizeInfo(dwInputStreamIndex: DWORD; out pcbSize, pcbMaxLookahead, pcbAlignment: DWORD): HRESULT; override;
    function InternalGetOutputSizeInfo(dwOutputStreamIndex: DWORD; out pcbSize, pcbAlignment: DWORD): HRESULT; override;
    function InternalGetInputMaxLatency(dwInputStreamIndex: DWORD; out prtMaxLatency: TREFERENCE_TIME): HRESULT; override;
    function InternalSetInputMaxLatency(dwInputStreamIndex: DWORD; rtMaxLatency: TREFERENCE_TIME): HRESULT; override;
    function InternalFlush: HRESULT; override;
    function InternalDiscontinuity(dwInputStreamIndex: DWORD): HRESULT; override;
    function InternalAllocateStreamingResources: HRESULT; override;
    function InternalFreeStreamingResources: HRESULT; override;
    function InternalProcessInput(dwInputStreamIndex: DWORD; pBuffer: IMediaBuffer; dwFlags: DWORD; rtTimestamp, rtTimelength: TREFERENCE_TIME): HRESULT; override;
    function InternalProcessOutput(dwFlags, cOutputBufferCount: DWORD; var pOutputBuffers: TDMO_OUTPUT_DATA_BUFFER_array; out pdwStatus: DWORD): HRESULT; override;
    function InternalAcceptingInput(dwInputStreamIndex: DWORD): HRESULT; override;
  public
    m_lVidWidth,                // Video width
    m_lVidHeight,               // Video Height
    m_lVidPitch: Longint;       // Video Pitch
    m_lVidSize: DWORD;          // Video Size
    destructor Destroy; override;
    procedure Initialize; override;
    procedure SetDirect3DDevice(D3DDevice: IDirect3DDevice8); stdcall;
    procedure GetDirect3DTexture(out Texture: IDirect3DTexture8); stdcall;
    procedure GetVideoDimentions(out Width, Height: DWORD); stdcall;
    property Marshal: IMarshal read fIMarshal implements IMarshal;
  end;

implementation

uses ComServ, ShellAPI, SysUtils, Registry;

function EqualGUID(const G1,G2: TGUID): Boolean;
begin
  Result := CompareMem(@G1, @G2, SizeOf(TGUID));
end;

{ TTextureRenderer }

procedure TTextureRenderer.Initialize;
var
  i: IUnknown;
begin
  fD3DDevice:= nil;
  fTexture:= nil;

  CoCreateFreeThreadedMarshaler(Self, i); //todo: Do we really need this ?
  fIMarshal:= i as IMarshal;

  NUMBEROFINPUTS:= 1;

  // Set NUMBEROFOUTPUTS to zero if you would like to simulate RendererFilter
  // !!!! But really DMO can't be used as Renderer !!!
  // Because: no syncro support, no EndOfStream notify, etc.

  // NUMBEROFOUTPUTS:= 0;
  NUMBEROFOUTPUTS:= 1;

  inherited;
end;

destructor TTextureRenderer.Destroy;
begin
  inherited;
end;

function TTextureRenderer.InternalAcceptingInput(
  dwInputStreamIndex: DWORD): HRESULT;
begin
  // Returning FALSE will result in GetInputStatus(...) returning
  //   pdwStatus := pdwStatus or DMO_INPUT_STATUSF_ACCEPT_DATA;
  Result:= S_OK;
  if (NUMBEROFOUTPUTS > 0) and Assigned(fMediaBuffer) then Result:= S_FALSE;
end;

procedure TTextureRenderer.InitD3DTexture;
var
  hr: HRESULT;
  ddsd: TD3DSurfaceDesc;
begin
  if Assigned(fTexture) or not Assigned(fD3DDevice) then Exit;

  // Create the texture that maps to this media type
  hr:= D3DXCreateTexture(fD3DDevice, m_lVidWidth, m_lVidHeight, 1, 0,
                         D3DFMT_A8R8G8B8, D3DPOOL_MANAGED, fTexture);
  if FAILED(hr) then
  begin
    OutputDebugString('Failed texture creatiion');
    Exit;
  end;

  // D3DXCreateTexture can silently change the parameters on us
  if FAILED(fTexture.GetLevelDesc(0, ddsd)) then
  begin
    OutputDebugString('Failed GetLevelDesc');
    fTexture:= nil;
    Exit;
  end;

  fTextureFormat:= ddsd.Format;
  if (fTextureFormat <> D3DFMT_A8R8G8B8) and
     (fTextureFormat <> D3DFMT_A1R5G5B5)
  then fTexture:= nil;
end;

function TTextureRenderer.InternalAllocateStreamingResources: HRESULT;
var
  bih: TBitmapInfoHeader;
begin
  Result:= S_OK;

  // Retrive the size of this media type
  Assert(IsEqualGUID(m_InputInfo[0].CurrentMediaType.formattype, FORMAT_VideoInfo));
  bih:= PVideoInfoHeader(m_InputInfo[0].CurrentMediaType.pbFormat).bmiHeader;

  m_lVidWidth  := bih.biWidth;
  m_lVidHeight := Abs(bih.biHeight);
  m_lVidPitch  := (m_lVidWidth * 3 + 3) and not(3); // We are forcing RGB24
  m_lVidSize   := bih.biSizeImage;

  InitD3DTexture;
end;

function TTextureRenderer.InternalFreeStreamingResources: HRESULT;
begin
  fTexture:= nil;
  Result:= S_OK;
end;

//  Helper - compares media types - ignoring the advisory fields
function TypesMatch(const pmt1, pmt2: PDMO_MEDIA_TYPE): Boolean;
begin
  if IsEqualGUID(pmt1.majortype,  pmt2.majortype) and
     IsEqualGUID(pmt1.subtype,    pmt2.subtype) and
     (pmt1.lSampleSize = pmt2.lSampleSize) and
     IsEqualGUID(pmt1.formattype, pmt2.formattype) and
     (pmt1.cbFormat = pmt2.cbFormat) and
     CompareMem(pmt1.pbFormat, pmt2.pbFormat, pmt1.cbFormat)
  then Result:= True
  else Result:= False;
end;

function TTextureRenderer.InternalCheckInputType(dwInputStreamIndex: DWORD;
  const pmt: PDMO_MEDIA_TYPE): HRESULT;
begin
  Result:= HRESULT(DMO_E_INVALIDTYPE);

  //  Check if the type is already set and if so reject any type that's not identical
  if (InputTypeSet(dwInputStreamIndex)) then
  begin
    if TypesMatch(pmt, InputType(dwInputStreamIndex)) then Result:= S_OK;
    Exit;
  end;

  //  We accept MEDIATYPE_Video, MEDIASUBTYPE_RGB24
  //  Check the format is defined
  if EqualGUID(pmt.majortype,  MEDIATYPE_Video) and
     EqualGUID(pmt.subtype,    MEDIASUBTYPE_RGB24) and
     EqualGUID(pmt.formattype, FORMAT_VideoInfo) and
     (pmt.pbFormat <> nil)
  then Result:= S_OK; // Finally approve it!!!
end;

function TTextureRenderer.InternalCheckOutputType(
  dwOutputStreamIndex: DWORD; const pmt: PDMO_MEDIA_TYPE): HRESULT;
var
  bihInput, bihOutput: PBitmapInfoHeader;
  rcOutputTarget: PRect;
  lWidth, lHeight: Longint;
begin
  Result:= HRESULT(DMO_E_INVALIDTYPE);

  //  Check if the type is already set and if so reject any type that's not identical
  if (OutputTypeSet(dwOutputStreamIndex)) then
  begin
    if (not TypesMatch(pmt, OutputType(dwOutputStreamIndex))) then Exit else
    begin
      Result:= S_OK;
      Exit;
    end;
  end;

  //  We output frames on stream 1
  if (dwOutputStreamIndex = 0) then
  begin
    if (not InputTypeSet(0)) then Exit;

    if EqualGUID(pmt.majortype,  MEDIATYPE_Video) and
       EqualGUID(pmt.subtype,    MEDIASUBTYPE_RGB24) and
       EqualGUID(pmt.formattype, FORMAT_VideoInfo) and
       (pmt.pbFormat <> nil) then
    begin
      bihOutput:= @PVideoInfoHeader(pmt.pbFormat).bmiHeader;
      rcOutputTarget:= @PVideoInfoHeader(pmt.pbFormat).rcTarget;

      Assert(IsEqualGUID(InputType(0).formattype, FORMAT_VideoInfo));
      bihInput:= @PVideoInfoHeader(InputType(0).pbFormat).bmiHeader;

      if (IsRectEmpty(rcOutputTarget^)) then
      begin
        lWidth := bihOutput.biWidth;
        lHeight := bihOutput.biHeight;
      end else
      begin
        lWidth := rcOutputTarget.right - rcOutputTarget.left;
        lHeight := rcOutputTarget.bottom - rcOutputTarget.top;
      end;

      if (bihInput.biWidth = lWidth) and
         (bihInput.biHeight = lHeight)
      then Result:= S_OK; // Finally here we approve it!!!
    end;
  end;
end;

function TTextureRenderer.InternalProcessInput(dwInputStreamIndex: DWORD;
  pBuffer: IMediaBuffer; dwFlags: DWORD; rtTimestamp,
  rtTimelength: TREFERENCE_TIME): HRESULT;
var
  Len: Cardinal;
  pBmpBuffer, pTxtBuffer: PByteArray; // Bitmap buffer, texture buffer
  lTxtPitch: Longint;                 // Pitch of bitmap, texture

  d3dlr: TD3DLockedRect;
  x, y: Integer;
  pBmpBufferOld: PByteArray;
  pTxtBufferOld: PByteArray;
begin
  if Assigned(fMediaBuffer) then // Wait until app cahlls ProcessOutput
  begin
    Result:= HResult(DMO_E_NOTACCEPTING);
    Exit;
  end;

  // Store current frame data
  if NUMBEROFOUTPUTS > 0 then fMediaBuffer:= pBuffer;
  fTimestamp:=  rtTimestamp;
  fTimelength:= rtTimelength;

  if not Assigned(fTexture) then
  begin
    Result := S_OK;
    Exit;
  end;

  // Get the video bitmap buffer
  pBuffer.GetBufferAndLength(@pBmpBuffer, @len);

  // Lock the Texture
  if (FAILED(fTexture.LockRect(0, d3dlr, nil, 0))) then
  begin
    Result := E_FAIL;
    Exit;
  end;

  // Get the texture buffer & pitch
  pTxtBuffer:= d3dlr.pBits;
  lTxtPitch:= d3dlr.Pitch;

  // Copy the bits
  // OPTIMIZATION OPPORTUNITY: Use a video and texture
  // format that allows a simpler copy than this one.
  if (fTextureFormat = D3DFMT_A8R8G8B8) then
  begin
    for y:= 0 to m_lVidHeight - 1 do
    begin
      pBmpBufferOld := pBmpBuffer;
      pTxtBufferOld := pTxtBuffer;
      for x:= 0 to m_lVidWidth - 1 do
      begin
        pTxtBuffer[0] := pBmpBuffer[0];
        pTxtBuffer[1] := pBmpBuffer[1];
        pTxtBuffer[2] := pBmpBuffer[2];
        pTxtBuffer[3] := $ff;
        Inc(Longint(pTxtBuffer), 4);
        Inc(Longint(pBmpBuffer), 3);
      end;

      Integer(pBmpBuffer):= Integer(pBmpBufferOld) + m_lVidPitch;
      Integer(pTxtBuffer):= Integer(pTxtBufferOld) + lTxtPitch;
    end;
  end;

  if (fTextureFormat = D3DFMT_A1R5G5B5) then
  begin
    for y:= 0 to m_lVidHeight - 1 do
    begin
      pBmpBufferOld := pBmpBuffer;
      pTxtBufferOld := pTxtBuffer;
      for x:= 0 to m_lVidWidth - 1 do
      begin
        PWord(@pTxtBuffer)^ :=
            $8000 +
            ((pBmpBuffer[2] and $F8) shl 7) +
            ((pBmpBuffer[1] and $F8) shl 2) +
            (pBmpBuffer[0] shr 3);

        Inc(Longint(pTxtBuffer), 2);
        Inc(Longint(pBmpBuffer), 3);
      end;
      Integer(pBmpBuffer):= Integer(pBmpBufferOld) + m_lVidPitch;
      Integer(pTxtBuffer):= Integer(pTxtBufferOld) + lTxtPitch;
    end;
  end;

  // Unlock the Texture
  if FAILED(fTexture.UnlockRect(0)) then
  begin
    Result := E_FAIL;
    Exit;
  end;

  Result:= S_OK;
  if (NUMBEROFOUTPUTS = 0) then
    Result:= S_FALSE; // no need for DMO wrapper to call ProcessOutput
end;

function TTextureRenderer.InternalProcessOutput(dwFlags,
  cOutputBufferCount: DWORD; var pOutputBuffers: TDMO_OUTPUT_DATA_BUFFER_array;
  out pdwStatus: DWORD): HRESULT;
var
  OutputBuffer: IMediaBuffer;
  Len1, Len2, ll: Cardinal;
  pSrc, pDst: PByteArray;
begin
  if NUMBEROFOUTPUTS = 0 then
  begin
    Result:= S_FALSE;
    Exit;
  end;

  if not Assigned(fMediaBuffer) then // Wait until app calls ProcessInput
  begin
    Result:= S_FALSE;
    Exit;
  end;

  pOutputBuffers[0].rtTimestamp:=  fTimestamp;
  pOutputBuffers[0].rtTimelength:= fTimelength;
  pOutputBuffers[0].dwStatus:=
    DMO_OUTPUT_DATA_BUFFERF_TIME or DMO_OUTPUT_DATA_BUFFERF_TIMELENGTH;

  OutputBuffer:= pOutputBuffers[0].pBuffer;

  fMediaBuffer.GetBufferAndLength(@pSrc, @Len1);
  OutputBuffer.GetMaxLength(Len2);

  Assert(m_lVidSize <= Len1);
  Assert(m_lVidSize <= Len2);

  OutputBuffer.SetLength(m_lVidSize);
  OutputBuffer.GetBufferAndLength(@pDst, @ll);

  Move(pSrc[0], pDst[0], Len2);
  fMediaBuffer:= nil;

  Result:= S_OK;
end;

function TTextureRenderer.InternalDiscontinuity(
  dwInputStreamIndex: DWORD): HRESULT;
begin
  Result:= S_OK;
end;

function TTextureRenderer.InternalFlush: HRESULT;
begin
  fMediaBuffer:= nil;
  Result:= S_OK;
end;

function TTextureRenderer.InternalGetInputMaxLatency(
  dwInputStreamIndex: DWORD; out prtMaxLatency: TREFERENCE_TIME): HRESULT;
begin
  Result:= E_NOTIMPL;
end;

function TTextureRenderer.InternalGetInputStreamInfo(
  dwInputStreamIndex: DWORD; out pdwFlags: DWORD): HRESULT;
begin
//  pdwFlags:= 0;
  pdwFlags:= DMO_INPUT_STREAMF_WHOLE_SAMPLES or
             DMO_INPUT_STREAMF_SINGLE_SAMPLE_PER_BUFFER or
             DMO_INPUT_STREAMF_FIXED_SAMPLE_SIZE;
  Result:= S_OK;
end;

function TTextureRenderer.InternalGetOutputStreamInfo(
  dwOutputStreamIndex: DWORD; out pdwFlags: DWORD): HRESULT;
begin
//  pdwFlags:= 0;
  pdwFlags:= DMO_INPUT_STREAMF_WHOLE_SAMPLES or
             DMO_INPUT_STREAMF_SINGLE_SAMPLE_PER_BUFFER or
             DMO_INPUT_STREAMF_FIXED_SAMPLE_SIZE;
  Result:= S_OK;
end;

function TTextureRenderer.InternalGetInputSizeInfo(
  dwInputStreamIndex: DWORD; out pcbSize, pcbMaxLookahead,
  pcbAlignment: DWORD): HRESULT;
begin
  pcbSize:= 1; // default implementation imposes no size requirements
  pcbMaxLookahead:= 4; // default implementation assumes no lookahead
//  pcbAlignment:= 1; // default implementation assumes no alignment
  pcbAlignment:= 4; // default implementation assumes no alignment
  Result:= S_OK;
end;

function TTextureRenderer.InternalGetOutputSizeInfo(
  dwOutputStreamIndex: DWORD; out pcbSize, pcbAlignment: DWORD): HRESULT;
var
  bih: TBitmapInfoHeader;
begin
  Assert(IsEqualGUID(m_InputInfo[0].CurrentMediaType.formattype, FORMAT_VideoInfo));
  bih:= PVideoInfoHeader(m_InputInfo[0].CurrentMediaType.pbFormat).bmiHeader;

  m_lVidWidth  := bih.biWidth;
  m_lVidHeight := Abs(bih.biHeight);
  m_lVidPitch  := (m_lVidWidth * 3 + 3) and not(3); // We are forcing RGB24

  if bih.biSizeImage <> 0 then pcbSize:= bih.biSizeImage
    else pcbSize:= m_lVidPitch*m_lVidHeight;

  pcbAlignment:= 4;
  Result:= S_OK;
end;

function TTextureRenderer.InternalGetInputType(dwInputStreamIndex,
  dwTypeIndex: DWORD; out pmt: TDMO_MEDIA_TYPE): HRESULT;
begin
  if (dwInputStreamIndex <> 0) or (dwTypeIndex <> 0) then
  begin //  We allowing only single input and one input format
    Result:= HResult(DMO_E_NO_MORE_ITEMS);
    Exit;
  end;

  // If GetInputType()'s pmt parameter is NULL, return S_OK if the type exists.
  // Return DMO_E_NO_MORE_ITEMS if the type does not exists.  See the
  // documentation for IMediaObject::GetInputType() for more information.
  if (nil <> @pmt) then
  begin
    //  Create our media type
    pmt.majortype:= MEDIATYPE_Video;
    pmt.subtype:= MEDIASUBTYPE_RGB24;
    pmt.bFixedSizeSamples    := TRUE;
    pmt.bTemporalCompression := FALSE;
    pmt.lSampleSize          := 0;
    pmt.formattype           := FORMAT_None;
    pmt.pUnk                 := nil;
  end;

  Result := S_OK;
end;

function TTextureRenderer.InternalGetOutputType(dwOutputStreamIndex,
  dwTypeIndex: DWORD; out pmt: TDMO_MEDIA_TYPE): HRESULT;
begin
  if (dwOutputStreamIndex <> 0) or (dwTypeIndex <> 0) then
  begin //  We allowing only single input and one input format
    Result:= HResult(DMO_E_NO_MORE_ITEMS);
    Exit;
  end;

  // If GetInputType()'s pmt parameter is NULL, return S_OK if the type exists.
  // Return DMO_E_NO_MORE_ITEMS if the type does not exists.  See the
  // documentation for IMediaObject::GetInputType() for more information.
  if not InputTypeSet(0) then
  begin
    if (nil <> @pmt) then
    begin
      //  Create our media type
      pmt.majortype:= MEDIATYPE_Video;
      pmt.subtype:= MEDIASUBTYPE_RGB24;
      pmt.bFixedSizeSamples    := TRUE;
      pmt.bTemporalCompression := FALSE;
      pmt.lSampleSize          := 0;
      pmt.formattype           := FORMAT_None;
      pmt.pUnk                 := nil;
    end;
  end else // Passthrow input data set (if it's already set)
  begin
    Result := MoCopyMediaType(pmt, @m_InputInfo[0].CurrentMediaType);
    if FAILED(Result) then Exit;
  end;

  Result:= S_OK;
end;

function TTextureRenderer.InternalSetInputMaxLatency(
  dwInputStreamIndex: DWORD; rtMaxLatency: TREFERENCE_TIME): HRESULT;
begin
  Result:= E_NOTIMPL;
end;

procedure TTextureRenderer.SetDirect3DDevice(D3DDevice: IDirect3DDevice8);
begin
  fD3DDevice:= D3DDevice;
  fTexture:= nil;
  InitD3DTexture;
end;

procedure TTextureRenderer.GetDirect3DTexture(
  out Texture: IDirect3DTexture8);
begin
  Texture:= fTexture;
end;

procedure TTextureRenderer.GetVideoDimentions(out Width, Height: DWORD);
begin
  Width:= m_lVidWidth;
  Height:= m_lVidHeight;
end;

type
  TTextureRendererFactory = class(TComObjectFactory)
  public
    procedure UpdateRegistry(Register: Boolean); override;
  end;

procedure TTextureRendererFactory.UpdateRegistry(Register: Boolean);
var
  mt: TDMO_PARTIAL_MEDIATYPE;
begin
  if Register then
  begin
    inherited UpdateRegistry(Register);

    // Register ourselves as a DMO with no types
    // RGS file also registers with a merit of 0 so DShow won't try to use us
    // But we also can register NO TYPES AT ALL
    mt.type_   := MEDIATYPE_Video;
    mt.subtype := MEDIASUBTYPE_RGB24;

    // If we use 0 instead of MERIT_NORMAL
    // when filter will not be visible in Filter Graph Editor (GraphEdit)
    DMORegister('_ DMO Texture Renderer', CLSID_TextureRenderer,
      DMOCATEGORY_VIDEO_DECODER, MERIT_NORMAL, 1, @mt, 0, nil);
      // DMOCATEGORY_VIDEO_EFFECT, MERIT_NORMAL, 1, @mt, 0, nil);
  end else
  begin
    DMOUnregister(CLSID_TextureRenderer, DMOCATEGORY_VIDEO_EFFECT);

    inherited UpdateRegistry(Register);
  end;
end;

initialization
  TTextureRendererFactory.Create(ComServer, TTextureRenderer,
    CLSID_TextureRenderer, '', 'Delphi DMO TextureRenderer DirectShow Filter',
    ciMultiInstance, tmApartment);
end.

