(*----------------------------------------------------------------------------*
 *  DirectShow and Direct3D sample for DirectX 8.x                            *
 *  by Alexey Barkovoy (E-Mail: clootie@reactor.ru)                           *
 *                                                                            *
 *  Modified: 14-May-2002                                                     *
 *                                                                            *
 *  This sample is modified Texture3D sample from DirectX8 SDK, what uses     *
 *  DirectShow for playing AVI and Direct3D8 for displaying it simultaneusly  *
 *  DMO filter model used in DirectShow part of example. So example shows how *
 *  to use helper class from DMOimpl (included in DirectShow.pas) in Delphi.  *
 *  Note: DMO - DirectX Media Objects                                         *
 *                                                                            *
 *  Also this sample shows how to use destination alpha to create transition  *
 *  effects what can be encoded in alpha-layer of image                       *
 *  (plasma.dds image is copyrighted by ATi technologies)                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru                                                *
 *----------------------------------------------------------------------------*
 *  *** Installation ***                                                      *
 *  This sample requries DirectX 8.1 to run. Also sample uses 'skiing.avi'    *
 *  from DirectX SDK (7 - 8.1), so if you do not have SDK instaled either     *
 *  copy and place in example directory another AVI with this name or download*
 *  it from: http://clootie.narod.ru/DelphiGraphics/Resources/skiing.zip      *
 *  Also D3DX8 wrapper DLL for DirectX 8.1 is required (d3dx81ab.dll),        *
 *  accessible at: http://clootie.narod.ru/DelphiGraphics/DX81/D3DX81_dll.zip *
 *----------------------------------------------------------------------------*)

unit Textures;

interface

uses
  Windows, Messages, SysUtils, MMSystem, ActiveX, ShellAPI, CommDlg,
  DirectShow, D3DX8,
  {$IFDEF DXG_COMPAT} DirectXGraphics {$ELSE} Direct3D8 {$ENDIF},
  D3DFont, DXUtil, DShow3DBridge;

const
  IDI_TEXTURES = 10000; // Icon resource handle
  FilterDLL = 'text3Ddmo.dll';
const
  SOURCE_FILE: String = 'skiing.AVI';
//  SOURCE_FILE: String = 'butterfly.mpg';

//-----------------------------------------------------------------------------
// Forward Declarations
//-----------------------------------------------------------------------------
function InitDShowTextureRenderer: HRESULT;
procedure CheckMovieStatus;
procedure CleanupDShow;
{$IFDEF REGISTER_FILTERGRAPH}
function AddToROT(const pUnkGraph: IUnknown): HRESULT;
procedure RemoveFromROT;
{$ENDIF}
procedure Msg(szFormat: PChar; Params: array of const);

procedure Textures_WinMain;

implementation

const
  CLASSNAME = 'DShow Texture3D Sample Class';

//-----------------------------------------------------------------------------
// Direct3D global variables
//-----------------------------------------------------------------------------
var
  hWindow: HWND;
  g_pD3D:       IDirect3D8;		 // Used to create the D3DDevice
  g_pd3dDevice: IDirect3DDevice8;        // Our rendering device
  g_pVB:        IDirect3DVertexBuffer8;	 // Buffer to hold vertices
  g_pVB2:       IDirect3DVertexBuffer8;	 // Buffer to hold vertices
  g_pTexture:   IDirect3DTexture8;       // Our texture
  m_pFont:      CD3DFont;                // Font for outputting frame stats
  VidWidth, VidHeight: DWORD;            // Video dimentions
  TexWidthAspect, TexHeightAspect: Single; // Aspect ratios for non-square videos
const
  g_bShowHelp:  Boolean         = False; // Show Help screen
  g_bShowPlasma:Boolean         = True;  // Show Plasma FX (alpha effect with AVI)

//-----------------------------------------------------------------------------
// Global DirectShow pointers
//-----------------------------------------------------------------------------
var
  g_pGB: IGraphBuilder;          // GraphBuilder
  g_pMC: IMediaControl;          // Media Control
  g_pMP: IMediaPosition;         // Media Postion
  g_pME: IMediaEvent;            // Media Event
  g_pMS: IMediaSeeking;
  pD3DDeviceExch: IDirect3DDeviceExch;
  CanDoDestinAlpha: Boolean = False;  // Can we use Destination alpha?
  IsOddPass: Boolean = False;
  TransTexture: IDirect3DTexture8;

type
// A structure for our custom vertex type. We added texture coordinates
  TColorTextureVertex = packed record
    pos: TD3DXVector3; // The position
    col: TD3DColor;    // The color
    tu, tv: Single;    // The texture coordinates
  end;

// Our custom FVF, which describes our custom vertex structure
const
  D3DFVF_CUSTOMVERTEX = (D3DFVF_XYZ or D3DFVF_DIFFUSE or D3DFVF_TEX1);

  AVIRect: array [0..3] of TColorTextureVertex = (
    (pos: (x:-1; y:-1; z:0); tu: 0; tv:0),
    (pos: (x:-1; y: 1; z:0); tu: 0; tv:1),
    (pos: (x: 1; y:-1; z:0); tu: 1; tv:0),
    (pos: (x: 1; y: 1; z:0); tu: 1; tv:1)
  );

var
  TransitionRect: array [0..3] of TColorTextureVertex = (
    (pos: (x:-1; y:-1; z:0); tu: 0; tv:1),
    (pos: (x:-1; y: 1; z:0); tu: 0; tv:0),
    (pos: (x: 1; y:-1; z:0); tu: 1; tv:1),
    (pos: (x: 1; y: 1; z:0); tu: 1; tv:0)
  );

// Register/Unregister Filter 
procedure RegisterFilters(Lib: PChar; Register: Boolean = True);
begin
  if Register then
    ShellExecute(0, 'open', 'regsvr32.exe', PChar('/s ' + Lib), nil, SW_HIDE)
  else
    ShellExecute(0, 'open', 'regsvr32.exe', PChar('/s /u ' + Lib), nil, SW_HIDE);
end;

procedure OpenFile;
var
  g_strFilename: array [0..511] of Char;
{$WRITEABLECONST ON}
const
  ofn: TOpenFilename = (
    lStructSize: SizeOf(TOpenFilename);
    hWndOwner: 0;
    hInstance: 0;
    lpstrFilter: 'AVI Files (*.avi)'#0'*.avi'#0'Any other video file (*.*)'#0'*.*'#0#0;
    lpstrCustomFilter: nil;
    nMaxCustFilter: 0;
    nFilterIndex: 1;
    lpstrFile: nil; // updated by code later
    nMaxFile: 512;
    lpstrFileTitle: nil; // updated by code later
    nMaxFileTitle: 512;
    lpstrInitialDir: nil; // updated by code later
    lpstrTitle: 'Open Mesh File';
    Flags: OFN_FILEMUSTEXIST;
    nFileOffset: 0;
    nFileExtension: 1;
    lpstrDefExt: nil;
    lCustData: 0;
    lpfnHook: nil;
    lpTemplateName: nil
  );
{$WRITEABLECONST OFF}
begin
  // Display the OpenFileName dialog. Then, try to load the specified file
  g_strFilename:= '';

  {$IFDEF COMPILER6_UP}
  // Handle Delphi6 case of extended OFN structure
  // (new members are invalid for Win98SE & below OS'es)
  ofn.lStructSize:= 76;

  {$ENDIF}

  StrCopy(g_strFilename, PChar(SOURCE_FILE));
  ofn.lpstrFile:= g_strFilename; 
  ofn.lpstrFileTitle:= nil;
  ofn.lpstrInitialDir:= nil;

  if GetOpenFileName(ofn)
  then SOURCE_FILE:= g_strFilename
  else SOURCE_FILE:= '';
end;

//-----------------------------------------------------------------------------
// Name: InitD3D()
// Desc: Initializes Direct3D
//-----------------------------------------------------------------------------
function InitD3D(hWnd: HWnd): HResult;
var
  d3ddm: TD3DDisplayMode;
  d3dpp: TD3DPresentParameters;
begin
  // Create the D3D object.
  g_pD3D:= Direct3DCreate8(D3D_SDK_VERSION);
  if (g_pD3D = nil) then
  begin
    Result:= E_FAIL;
    Exit;
  end;

  // Get the current desktop display mode, so we can set up a back
  // buffer of the same format
  Result := g_pD3D.GetAdapterDisplayMode(D3DADAPTER_DEFAULT, d3ddm);
  if FAILED(Result) then
  begin
    Msg('Could not read adapter display mode!  hr:=0x%x', [Result]);
    Exit;
  end;

  if (d3ddm.Format = D3DFMT_X8R8G8B8) then
  begin // try to Create Back Buffer with destination alpha
    if not FAILED(g_pD3D.CheckDeviceType(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL,
                           D3DFMT_X8R8G8B8, D3DFMT_A8R8G8B8, True)) then
    begin
      d3ddm.Format:= D3DFMT_A8R8G8B8;
      CanDoDestinAlpha:= True;
    end;
  end;

  // Set up the structure used to create the D3DDevice. Since we are now
  // using more complex geometry, we will create a device with a zbuffer.
  FillChar(d3dpp, SizeOf(d3dpp), 0);
  d3dpp.Windowed               := TRUE;
  d3dpp.SwapEffect             := D3DSWAPEFFECT_COPY_VSYNC;
  d3dpp.BackBufferFormat       := d3ddm.Format;
  d3dpp.EnableAutoDepthStencil := TRUE;
  d3dpp.AutoDepthStencilFormat := D3DFMT_D16;

  // Create the D3DDevice
  Result := g_pD3D.CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd,
               D3DCREATE_SOFTWARE_VERTEXPROCESSING or D3DCREATE_MULTITHREADED,
               d3dpp, g_pd3dDevice);
  if FAILED(Result) then
  begin
    Msg('Could not create the D3D device!  hr:=0x%x'#13#13 +
        'This sample is attempting to create a buffer that might not'#13 +
        'be supported by your video card in its current mode.'#13#13 +
        'You may want to reduce your screen resolution or bit depth'#13 +
        'and try to run this sample again.', [Result]);
    Exit;
  end;

  // Turn off culling
  g_pd3dDevice.SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);

  // Turn off D3D lighting
  g_pd3dDevice.SetRenderState(D3DRS_LIGHTING, iFalse);

  // Turn on the zbuffer
  g_pd3dDevice.SetRenderState(D3DRS_ZENABLE, iTrue);

  // Create simple Font object
  m_pFont := CD3DFont.Create('Arial', 12, D3DFONT_BOLD);
  m_pFont.InitDeviceObjects(g_pd3dDevice);
  m_pFont.RestoreDeviceObjects;

  // Load "Transition" texture if we have Destination alpha support
  if CanDoDestinAlpha then
  begin
    if Failed(D3DXCreateTextureFromFile(g_pd3dDevice,
      PChar(ExtractFilePath(ParamStr(0)) + '\' + 'plasma.dds'), TransTexture)) then
      TransTexture:= nil;
  end;
end;


//-----------------------------------------------------------------------------
// Name: InitGeometry()
// Desc: Create the textures and vertex buffers
//-----------------------------------------------------------------------------
function InitGeometry: HResult;
type
  PCUSTOMVERTEXarray = ^TCUSTOMVERTEXarray;
  TCUSTOMVERTEXarray = array [0..0] of TColorTextureVertex;
var
  pVertices: PCUSTOMVERTEXarray;
  pVertices2: PCUSTOMVERTEXarray;
  i: DWord;
  theta: Single;
  SurfDesc: TD3DSurfaceDesc;
begin
  // Create the vertex buffers
  Result := g_pd3dDevice.CreateVertexBuffer(50*2*SizeOf(TColorTextureVertex),
                           0, D3DFVF_CUSTOMVERTEX, D3DPOOL_DEFAULT, g_pVB);
  if FAILED(Result) then
  begin
    Msg('Could not create a vertex buffer!  hr:=0x%x', [Result]);
    Result := E_FAIL;
    Exit;
  end;

  Result := g_pd3dDevice.CreateVertexBuffer(50*2*SizeOf(TColorTextureVertex),
                           0, D3DFVF_CUSTOMVERTEX, D3DPOOL_DEFAULT, g_pVB2);
  if FAILED(Result) then
  begin
    Msg('Could not create a vertex buffer!  hr:=0x%x', [Result]);
    Result := E_FAIL;
    Exit;
  end;

  // Fill the vertex buffer. We are setting the tu and tv texture
  // coordinates, which range from 0.0 to 1.0
  Result := g_pVB.Lock(0, 0, PByte(pVertices), 0);
  if FAILED(Result) then
  begin
    Msg('Could not lock the vertex buffer!  hr:=0x%x', [Result]);
    Result := E_FAIL;
    Exit;
  end;
  Result := g_pVB2.Lock(0, 0, PByte(pVertices2), 0);
  if FAILED(Result) then
  begin
    Msg('Could not lock the vertex buffer!  hr:=0x%x', [Result]);
    Result := E_FAIL;
    Exit;
  end;

  // Init texture aspect ratios
  pD3DDeviceExch.GetVideoDimentions(VidWidth, VidHeight);
  g_pTexture.GetLevelDesc(0, SurfDesc);
  TexWidthAspect:=  VidWidth  / SurfDesc.Width;
  TexHeightAspect:= VidHeight / SurfDesc.Height;

  for i:= 0 to 50 - 1 do
  begin
    theta := (2*D3DX_PI*i)/(50-1);

    pVertices[2*i+0].pos := D3DXVector3(Sin(theta), -1.0, Cos(theta));
    pVertices[2*i+0].col := D3DCOLOR_ARGB(0, $FF,$FF,$FF);
    pVertices[2*i+0].tu  := (TexWidthAspect*i)/(50-1);
    pVertices[2*i+0].tv  := TexHeightAspect;

    pVertices[2*i+1].pos := D3DXVector3(Sin(theta), 1.0, Cos(theta));
    pVertices[2*i+1].col := D3DCOLOR_ARGB(0, $80,$80,$80);
    pVertices[2*i+1].tu  := (TexWidthAspect*i)/(50-1);
    pVertices[2*i+1].tv  := 0.0;

    pVertices2[2*i+0].pos := D3DXVector3(Sin(theta), -1.0, Cos(theta));
    pVertices2[2*i+0].col := D3DCOLOR_ARGB(0, $80,$80,$FF);
    pVertices2[2*i+0].tu  := (TexWidthAspect*i)/(50-1);
    pVertices2[2*i+0].tv  := TexHeightAspect;

    pVertices2[2*i+1].pos := D3DXVector3(Sin(theta), 1.0, Cos(theta));
    pVertices2[2*i+1].col := D3DCOLOR_ARGB(0, $0,$0,$80);
    pVertices2[2*i+1].tu  := (TexWidthAspect*i)/(50-1);
    pVertices2[2*i+1].tv  := 0.0;
  end;

  g_pVB2.Unlock;
  g_pVB.Unlock;
  Result := S_OK;
end;



//-----------------------------------------------------------------------------
// Name: Cleanup()
// Desc: Releases all previously initialized objects
//-----------------------------------------------------------------------------
procedure Cleanup;
begin
  CleanupDShow;

  m_pFont.Free;
  g_pTexture:= nil;
  g_pVB2:= nil;
  g_pVB:= nil;
  g_pd3dDevice:= nil;
  g_pD3D:= nil;
end;

var
  OldVWVisible: LongBool = False;

procedure SwitchActiveMovieWindow(DoRestore: Boolean = False);
var
  VideoWindow: IVideoWindow;
begin
  VideoWindow:= g_pGB as IVideoWindow;
  if DoRestore then
  begin
    VideoWindow.put_Visible(OldVWVisible);
  end else
  begin
    if Succeeded(VideoWindow.get_Visible(OldVWVisible)) then
    begin
      OldVWVisible:= not OldVWVisible;
      VideoWindow.put_Visible(OldVWVisible);
    end;
  end;
  
  // Set Focus back to our main window
  if OldVWVisible then SetFocus(hWindow)
end;


//-----------------------------------------------------------------------------
// Name: SetupMatrices()
// Desc: Sets up the world, view, and projection transform matrices.
//-----------------------------------------------------------------------------
procedure SetupMatrices(NearPlane: Single = 2);
var
  Result: HResult;
  matWorld: TD3DXMatrix;
  matView: TD3DXMatrix;
  matProj: TD3DXMatrix;
begin
  // For our world matrix, we will just leave it as the identity
  D3DXMatrixIdentity(matWorld);
  D3DXMatrixRotationX(matWorld, timeGetTime/2000.0);
  Result := g_pd3dDevice.SetTransform(D3DTS_WORLD, matWorld);
  if FAILED(Result) then
  begin
    Msg('Could not set D3DTS_WORLD transform!  hr:=0x%x', [Result]);
  end;

  // Set up our view matrix. A view matrix can be defined given an eye point,
  // a po : integer to lookat, and a direction for which way is up. Here, we set the
  // eye five units back along the z-axis and up three units, look at the
  // origin, and define up to be in the y-direction.
  D3DXMatrixLookAtLH(matView, D3DXVector3(0.0, 3.0,-5.0),
                              D3DXVector3(0.0, 0.0, 0.0),
                              D3DXVector3(0.0, 1.0, 0.0));
  Result := g_pd3dDevice.SetTransform(D3DTS_VIEW, matView);
  if FAILED(Result) then
  begin
    Msg('Could not set D3DTS_VIEW transform!  hr:=0x%x', [Result]);
  end;

  // For the projection matrix, we set up a perspective transform (which
  // transforms geometry from 3D view space to 2D viewport space, with
  // a perspective divide making objects smaller in the distance). To build
  // a perpsective transform, we need the field of view (1/4 pi is common),
  // the aspect ratio, and the near and far clipping planes (which define at
  // what distances geometry should be no longer be rendered).
  D3DXMatrixPerspectiveFovLH(matProj, D3DX_PI/4, 1.0, NearPlane, 100.0);
  Result := g_pd3dDevice.SetTransform(D3DTS_PROJECTION, matProj);
  if FAILED(Result) then
  begin
    Msg('Could not set D3DTS_PROJECTION transform!  hr:=0x%x', [Result]);
  end;
end;

// Set all transform matrices to identity
procedure SetupTransitionMatricesAndData;
var
  m: TD3DXMatrix;
  i: Integer;
  SomeTicks: Cardinal;
  a: Byte;
begin
  D3DXMatrixIdentity(m);
  g_pd3dDevice.SetTransform(D3DTS_PROJECTION, m);
  g_pd3dDevice.SetTransform(D3DTS_WORLD, m);
  g_pd3dDevice.SetTransform(D3DTS_VIEW, m);

  SomeTicks:= Round(timeGetTime/30);
  a:= SomeTicks mod ($FF+1);
  IsOddPass:= Odd(SomeTicks div ($FF+1));
  if not IsOddPass then a:= $FF - a; 

  for i:= 0 to 3 do
    TransitionRect[i].col:= D3DCOLOR_ARGB(a, 0,0,0);

  AVIRect[1].tv:= TexHeightAspect;
  AVIRect[2].tu:= TexWidthAspect;
  AVIRect[3].tu:= TexWidthAspect;
  AVIRect[3].tv:= TexHeightAspect;
end;

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Draws the scene
//-----------------------------------------------------------------------------
procedure Render;
begin
  // Clear the backbuffer and the zbuffer
  g_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER,
                     D3DCOLOR_ARGB(0,0,0,255), 1.0, 0);

  // Begin the scene
  g_pd3dDevice.BeginScene;

  // Common states
  g_pd3dDevice.SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
  g_pd3dDevice.SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
  g_pd3dDevice.SetTextureStageState(1, D3DTSS_COLOROP,   D3DTOP_DISABLE);
  g_pd3dDevice.SetTextureStageState(1, D3DTSS_ALPHAOP,   D3DTOP_DISABLE);

  g_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_CURRENT);
  g_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_TEXTURE);
  g_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_CURRENT);
  g_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAARG2, D3DTA_TEXTURE);

  g_pd3dDevice.SetVertexShader(D3DFVF_CUSTOMVERTEX);

  // Setup the world, view, and projection matrices
  SetupMatrices;

  g_pd3dDevice.SetRenderState(D3DRS_ZFUNC, D3DCMP_LESSEQUAL);
  g_pd3dDevice.SetRenderState(D3DRS_ZWRITEENABLE, iTrue);
  g_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iFalse);

  g_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_MODULATE);
  g_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAOP,   D3DTOP_MODULATE);

  // Render cylinder with AVI texture on it
  g_pd3dDevice.SetTexture(0, g_pTexture);
  g_pd3dDevice.SetStreamSource(0, g_pVB, SizeOf(TColorTextureVertex));
  g_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2*50-2);

  // Draw plasma FX
  if CanDoDestinAlpha and g_bShowPlasma then
  begin
    // Common states for planar FX
    SetupTransitionMatricesAndData;
    g_pd3dDevice.SetRenderState(D3DRS_ZWRITEENABLE, iFalse);

    //////////////////////////////////////////////////////////
    // Draw transition mask to back buffer
    g_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iTrue);
    g_pd3dDevice.SetRenderState(D3DRS_ALPHAFUNC, D3DCMP_GREATEREQUAL);
    g_pd3dDevice.SetRenderState(D3DRS_ALPHATESTENABLE, iTrue);
    g_pd3dDevice.SetRenderState(D3DRS_ALPHAREF, $000000E0);

    g_pd3dDevice.SetRenderState(D3DRS_SRCBLEND,  D3DBLEND_SRCCOLOR);
    g_pd3dDevice.SetRenderState(D3DRS_DESTBLEND, D3DBLEND_DESTCOLOR);

    g_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_ADD);
    g_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_ADD);

    if not IsOddPass
    then g_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAARG2, D3DTA_TEXTURE)
    else g_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAARG2, D3DTA_COMPLEMENT or D3DTA_TEXTURE);

    g_pd3dDevice.SetTexture(0, TransTexture);
    g_pd3dDevice.DrawPrimitiveUP(D3DPT_TRIANGLESTRIP, 2, TransitionRect, SizeOf(TColorTextureVertex));

    g_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAARG2, D3DTA_TEXTURE);
    //////////////////////////////////////////////////////////
    // Draw Original - not geometry-morfed AVI file picture
    g_pd3dDevice.SetRenderState(D3DRS_ALPHATESTENABLE, iFalse);

    g_pd3dDevice.SetRenderState(D3DRS_SRCBLEND,  D3DBLEND_DESTALPHA);
    g_pd3dDevice.SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVDESTALPHA);

    g_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_SELECTARG2);
    g_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAOP,   D3DTOP_SELECTARG2);

    g_pd3dDevice.SetTexture(0, g_pTexture);
    g_pd3dDevice.DrawPrimitiveUP(D3DPT_TRIANGLESTRIP, 2, AVIRect, SizeOf(TColorTextureVertex));

    //////////////////////////////////////////////////////////
    // Render BLUE colored cylinder
    g_pd3dDevice.SetRenderState(D3DRS_ZWRITEENABLE, iTrue);

    g_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_SELECTARG1);
    g_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAOP,   D3DTOP_SELECTARG1);

    // Setup the world, view, and projection matrices
    // Do some matrix tricks to aviod using "D3DRS_ZBIAS" render state
    SetupMatrices(2.01);

    g_pd3dDevice.SetTexture(0, nil);
    g_pd3dDevice.SetStreamSource(0, g_pVB2, SizeOf(TColorTextureVertex));
    g_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2*50-2);
  end;

  g_pd3dDevice.SetTexture(0, nil);

  if not g_bShowHelp then
  begin
    m_pFont.DrawText(2,  0, D3DCOLOR_ARGB(255,255,255,255), 'H');
    m_pFont.DrawText(20, 0, D3DCOLOR_ARGB(255,255,255,0), '- Press to display help screen');
  end else
  begin
    m_pFont.DrawText(2,  0, D3DCOLOR_ARGB(255,255,255,255), 'H');
    m_pFont.DrawText(20, 0, D3DCOLOR_ARGB(255,255,255,0), '- Press to hide help screen');

    m_pFont.DrawText(2,  24, D3DCOLOR_ARGB(255,255,255,255),
      'R'#10'P'#10'S'#10'F'#10'V'#10'O'#10'U');
    m_pFont.DrawText(20, 24, D3DCOLOR_ARGB(255,255,255,0),
      '- Run'#10 +
      '- Pause'#10 +
      '- Stop'#10 +
      '- Switch plasma FX on/off'#10 +
      '- Show/hide Renderer VideoWindow'#10 +
      '- Open another file'#10 +
      '- Uninstall Delphi DMO filter');
  end;

  // End the scene
  g_pd3dDevice.EndScene;

  // Present the backbuffer contents to the display
  g_pd3dDevice.Present(nil, nil, 0, nil);

  // Check to see if we need to restart the movie
  CheckMovieStatus;
end;

//-----------------------------------------------------------------------------
// InitDShowTextureRenderer : Create DirectShow filter graph and run the graph
//-----------------------------------------------------------------------------
function InitDShowTextureRenderer: HRESULT;
 function FiledConnect(const OutFilter: IBaseFilter; OutPinName: PWideChar;
   const InFilter: IBaseFilter; InPinName: PWideChar): Boolean;
 var
   hr: HResult;
   iPinIn, iPinOut: IPin;
{   i: IPin;
   ep: IEnumPins;
   l: Integer;
   pi: TPin_Info; }
 begin
   Result:= True;

   hr:= InFilter.FindPin(InPinName, iPinIn);
   if FAILED(hr) then
   begin
     Msg('Could not find input pin!  hr:=0x%x'#13'Pin: "%s"', [hr, InPinName]);
{     InFilter.EnumPins(ep);
     while ep.Next(1, i, @l) <> S_FALSE do
     begin
       i.QueryPinInfo(pi);
       Msg(@Char(pi.achName[0]), []);
     end; }
     Exit;
   end;

   hr:= OutFilter.FindPin(OutPinName, iPinOut);
   if FAILED(hr) then
   begin
     Msg('Could not find output pin!  hr:=0x%x'#13'Pin: "%s"', [hr, OutPinName]);
     Exit;
   end;

   hr:= g_pGB.Connect(iPinOut, iPinIn); // Connect these two filters
   if FAILED(hr) then
   begin
     Msg('Could not connect pins!  hr:=0x%x'#13'Pins: "%s", "%s"', [hr, OutPinName, InPinName]);
     Exit;
   end;

   Result:= False; // if we are here -> All OK
 end;

var
  pFTR: IBaseFilter;        // Texture Renderer Filter
//  pVideo: IBaseFilter; // Video renderer

  strFileName: String;
  wFileName: array[0..MAX_PATH] of WideChar;

  pFilter: IBaseFilter;
  pWrap: IDMOWrapperFilter;
begin
  // Create the filter graph
  Result:= CoCreateInstance(CLSID_FilterGraph, nil, CLSCTX_INPROC, IID_IGraphBuilder, g_pGB);
  if FAILED(Result) then
  begin
    Result := E_FAIL;
    Exit;
  end;

  {$IFDEF REGISTER_FILTERGRAPH}
  // Register the graph in the Running Object Table (for debug purposes)
  AddToROT(g_pGB);
  {$ENDIF}

  ////////////////////////////////////////
  // IMediaObject

  // Create the DMO Wrapper filter.
  Result := CoCreateInstance(CLSID_DMOWrapperFilter, nil,
      CLSCTX_INPROC, IID_IBaseFilter, pFilter);
  if FAILED(Result) then
  begin
    Msg('Could not CreateInstance DMO Wrapper interface!  hr:=0x%x', [Result]);
    Exit;
  end;

  Result := pFilter.QueryInterface(IID_IDMOWrapperFilter, pWrap);
  if FAILED(Result) then
  begin // Initialize the filter.
    Msg('Could not request DMO Wrapper filter!  hr:=0x%x', [Result]);
    Exit;
  end;

  Result := pWrap.Init(CLSID_TextureRenderer, DMOCATEGORY_VIDEO_EFFECT);
  if FAILED(Result) then  // Add the filter to the graph.
  begin
    Msg('Could not request DMO texture Renderer filter!  hr:=0x%x'#10 +
      'Trying to register it. Please, try to re-run demo once more.', [Result]);
    RegisterFilters(FilterDLL);
    Exit;
  end;
  pWrap:= nil;

  Result := pFilter.QueryInterface(IDirect3DDeviceExch, pD3DDeviceExch);
  if FAILED(Result) then
  begin // Initialize the filter.
    Msg('Could not request IDirect3DDeviceExch!  hr:=0x%x', [Result]);
    Exit;
  end;

  Result := g_pGB.AddFilter(pFilter, 'My DMO');
  if FAILED(Result) then
  begin
    Msg('xxxxxxx!  hr:=0x%x', [Result]);
    Exit;
  end;

  pFTR:= pFilter;
  pFilter:= nil;

  if not FileExists(SOURCE_FILE) then
  begin
    // Determine the file to load based on DirectX Media path (from SDK)
    // Use a helper function included in DXUtils.cpp
    SetLength(strFileName, StrLen(DXUtil_GetDXSDKMediaPath));
    StrCopy(PChar(strFileName), DXUtil_GetDXSDKMediaPath);
    SOURCE_FILE:= strFileName + SOURCE_FILE;
  end;

  while not FileExists(SOURCE_FILE) do
  begin
    MessageBox(0, PChar(Format('File does not exists:'#10'"%s"', [SOURCE_FILE])),
      'DirectShow Texture3D Sample', MB_OK or MB_ICONERROR);
    OpenFile;
    if (SOURCE_FILE = '') then // User cancelled file choosing
    begin
      Result:= E_FAIL;
      Exit;
    end;
  end;
  strFileName:= SOURCE_FILE;

  {$IFNDEF UNICODE}
  MultiByteToWideChar(CP_ACP, 0, PChar(strFileName), -1, wFileName, MAX_PATH);
  {$ELSE}
  lstrcpy(wFileName, strFileName);
  {$ENDIF}

  Result:= g_pGB.RenderFile(wFileName, nil);
  if FAILED(Result) then
  begin
    Msg('Failed "RenderFile"! hr:=0x%x', [Result]);
    Exit;
  end;

{
  // Add the source filter
  Result := g_pGB.AddSourceFilter (wFileName, 'SOURCE', pFSrc);
  if FAILED(Result) then
  begin
    Msg('Could not create source filter to graph!  hr:=0x%x', [Result]);
    Exit;
  end;

  if FiledConnect(pFSrc, 'Output', pFTR, 'in0') then Exit;

  Result:= CoCreateInstance(CLSID_VideoRenderer, nil, CLSCTX_INPROC, IID_IBaseFilter, pVideo);
  if FAILED(Result) then
  begin
    Result := E_FAIL;
    Exit;
  end;

  Result := g_pGB.AddFilter(pVideo, 'Video Renderer');
  if FAILED(Result) then
  begin
    Msg('Could not add Video Renderer filter to graph!  hr:=0x%x', [Result]);
    Exit;
  end;

  if FiledConnect(pFTR, 'out0', pVideo, 'Input') then Exit;
 }

  // Get the graph's media control, event andposition interfaces
  g_pMC:= g_pGB as IMediaControl;
  g_pMP:= g_pGB as IMediaPosition;
  g_pME:= g_pGB as IMediaEvent;
  g_pMS:= g_pGB as IMediaSeeking;

  // Force IVideoWindow to not show itself on start of movie
  (g_pGB as IVideoWindow).put_AutoShow(False);
  SwitchActiveMovieWindow(True);

  // Start the graph running;
  Result := g_pMC.Run;
  if FAILED(Result) then
  begin
    Msg('Could not run the DirectShow graph!  hr:=0x%x', [Result]);
    Exit;
  end;

  Result := S_OK;
end;


//-----------------------------------------------------------------------------
// CheckMovieStatus: If the movie has ended, rewind to beginning
//-----------------------------------------------------------------------------
procedure CheckMovieStatus;
var
  lEventCode:   Longint;
  lParam1:      Longint;
  lParam2:      Longint;
begin
  // Check for completion events
  g_pME.GetEvent(lEventCode, lParam1, lParam2, 0);
  if (EC_COMPLETE = lEventCode) then
    g_pMP.put_CurrentPosition(0);
  g_pME.FreeEventParams(lEventCode, lParam1, lParam2);
end;


//-----------------------------------------------------------------------------
// CleanupDShow
//-----------------------------------------------------------------------------
procedure CleanupDShow;
begin
  {$IFDEF REGISTER_FILTERGRAPH}
  // Pull graph from Running Object Table (Debug)
  RemoveFromROT;
  {$ENDIF}

  // Shut down the graph
  if Assigned(pD3DDeviceExch) then pD3DDeviceExch:= nil;

  if Assigned(g_pMC) then
  begin
    g_pMC.Stop;
    g_pMC:= nil;
  end;
  g_pMS:= nil;
  g_pGB:= nil;
end;

{$IFDEF REGISTER_FILTERGRAPH}

//-----------------------------------------------------------------------------
// Running Object Table functions: Used to debug. By registering the graph
// in the running object table, GraphEdit is able to connect to the running
// graph. This code should be removed before the application is shipped in
// order to avoid third parties from spying on your graph.
//-----------------------------------------------------------------------------
var
  dwROTReg: DWORD = $fedcba98;

function AddToROT(const pUnkGraph: IUnknown): HResult;
var
  pmk: IMoniker;
  pirot: IRunningObjectTable;
begin
  if (FAILED(GetRunningObjectTable(0, pirot))) then
  begin
    Result := E_FAIL;
    Exit;
  end;
  WCHAR wsz : array[ 0..256- 1 ] of ;
  wsprintfW(wsz, LFilterGraph %08x  pid %08x, (DWORD_PTR) 0, GetCurrentProcessId);
  HRESULT hr := CreateItemMoniker(L!, wsz, @pmk);
  if (SUCCEEDED(hr)) begin
      hr := pirot.Register(0, pUnkGraph, pmk, @dwROTReg);
      pmk.Release;
  end;
  pirot.Release;
  result := hr;
end;

procedure RemoveFromROT;
var
  pirot: IRunningObjectTable;
begin
  if SUCCEEDED(GetRunningObjectTable(0, @pirot))
  then pirot.Revoke(dwROTReg);
end;

{$ENDIF}


//-----------------------------------------------------------------------------
// Msg: Display an error message box if needed
//-----------------------------------------------------------------------------
procedure Msg(szFormat: PChar; Params: array of const);
var
  S: String;
begin
  MessageBox(0, PChar(Format(szFormat, Params)), 'DirectShow Texture3D Sample',
    MB_OK or MB_ICONERROR);

  SetLength(S, 255);
  AMGetErrorText(params[0].VInteger, PChar(S), 255);
  SetLength(S, StrLen(PChar(S)));
  MessageBox(0, PChar(S), 'DirectShow Texture3D Sample',
    MB_OK or MB_ICONERROR);
end;

//=============================================================================
//=============================================================================
//=============================================================================
//=============================================================================
//=============================================================================


//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: The window's message handler
//-----------------------------------------------------------------------------
function MsgProc(hWnd: HWND; msg: UINT; wParam: WPARAM; lParam: LPARAM ): LRESULT; stdcall;
var
  mcState: TFilter_State;
begin
  case msg of
    WM_DESTROY:
    begin
      PostQuitMessage(0);
      Result := 0;
      Exit;
    end;
    WM_Char:
    begin
      case Char(wParam) of
        'r', 'R':
        begin
          g_pMC.GetState(INFINITE, mcState);
          g_pMC.Run;
          if (mcState = State_Stopped) then
          begin
            // Get New updated texture interface
            pD3DDeviceExch.GetDirect3DTexture(g_pTexture);
          end;
        end;
        'p', 'P': g_pMC.Pause;
        's', 'S': g_pMC.Stop;
        'h', 'H': g_bShowHelp:= not g_bShowHelp;
        'v', 'V': SwitchActiveMovieWindow;
        'u', 'U': RegisterFilters(FilterDLL, False);
        'f', 'F':
        begin
          g_bShowPlasma:= not g_bShowPlasma;
          if not CanDoDestinAlpha then
            MessageBox(0, 'Sorry, plasma FX is not available.'#10 +
              'Your video doesn''t support destination alpha.',
              'DirectShow Texture3D Sample', MB_OK or MB_ICONERROR);
        end;
        'o', 'O':
        begin
          OpenFile;
          if SOURCE_FILE = '' then
          begin
            Result := 0;
            Exit;
          end;
          CleanupDShow;
          InitDShowTextureRenderer;

          // Receive our private interface to D3D8 texture
          pD3DDeviceExch.SetDirect3DDevice(g_pd3dDevice);
          pD3DDeviceExch.GetDirect3DTexture(g_pTexture);

          // Create the scene geometry
          InitGeometry;
        end;
      end;
    end;
  end;

  Result:= DefWindowProc(hWnd, msg, wParam, lParam);
end;


//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: The application's entry point
//-----------------------------------------------------------------------------
procedure Textures_WinMain;
{$WRITEABLECONST ON}
const
  wc: WNDCLASSEX = (style: CS_CLASSDC; lpfnWndProc: @MsgProc;
                    cbClsExtra: 0; cbWndExtra: 0;
                    hCursor: 0; hbrBackground: 0; lpszMenuName: nil;
                    lpszClassName: CLASSNAME; hIconSm: 0);
{$WRITEABLECONST OFF}
var
  dwWindowStyle: DWORD;
  rc: TRect;
  msg: tagMSG;
  WndWidth, WndHeight: DWORD;
begin
  // Initialize COM
  CoInitialize(nil);

  // Register the window class
  wc.cbSize:= SizeOf(WNDCLASSEX);
  wc.hInstance := HInstance;
  wc.hIcon := LoadIcon(hInstance, MAKEINTRESOURCE(IDI_TEXTURES));

  RegisterClassEx(wc);

try
  // DShow: Set up filter graph with our custom renderer.
  if FAILED(InitDShowTextureRenderer) then
  begin
    MessageBox(0, 'Error initializing DirectShow DMO Filter',
      'DirectShow Texture3D Sample', MB_OK or MB_ICONERROR);
    Exit;
  end;

  // Get video clip dimentions 
  pD3DDeviceExch.GetVideoDimentions(VidWidth, VidHeight);
  WndWidth:= VidWidth * 2;
  WndHeight:= VidHeight * 2;

  // Set the window's initial style
  dwWindowStyle := WS_OVERLAPPED or WS_CAPTION or WS_SYSMENU or WS_THICKFRAME or
                   WS_VISIBLE or WS_MINIMIZEBOX or WS_MAXIMIZEBOX;  

  // Set the window's initial width
  SetRect(rc, 0, 0, WndWidth, WndHeight);
  AdjustWindowRect(rc, dwWindowStyle, False);

  // Create the application's window
  hWindow := CreateWindow(CLASSNAME, 'DShow DMO Texture3D Sample', dwWindowStyle,
                          Integer(CW_USEDEFAULT), Integer(CW_USEDEFAULT),
                          (rc.right-rc.left), (rc.bottom-rc.top),
                          GetDesktopWindow, 0, wc.hInstance, nil);

  // Initialize Direct3D
  if SUCCEEDED(InitD3D(hWindow)) then
  begin
    // Receive our private interface to D3D8 texture
    pD3DDeviceExch.SetDirect3DDevice(g_pd3dDevice);
    pD3DDeviceExch.GetDirect3DTexture(g_pTexture);

    // Create the scene geometry
    if SUCCEEDED(InitGeometry) then
    begin
      // Show the window
      ShowWindow(hWindow, SW_SHOWDEFAULT);
      UpdateWindow(hWindow);

      // Enter the message loop
      FillChar(msg, SizeOf(msg), 0);
      while (msg.message <> WM_QUIT) do
      begin
        if PeekMessage(msg, 0, 0, 0, PM_REMOVE) then
        begin
          TranslateMessage(msg);
          DispatchMessage(msg);
        end
        else
          Render;
      end;
    end;
  end;

finally
  // Clean up everything and exit the app
  Cleanup;
  UnregisterClass(CLASSNAME, wc.hInstance);
  CoUninitialize;
end;
end;

end.

